/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "i386ArchProcessor.h"
#include "QCatagoryList.h"
#include "Debugger.h"
#include "DebuggerCoreInterface.h"
#include "DisassemblyLib.h"
#include "State.h"
#include "Util.h"

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
i386ArchProcessor::i386ArchProcessor() {

	// initialize our system with some libc definitions
	m_FunctionInfo.insert("memchr", FunctionInfo("memchr", 'p', 'c', 'u'));
	m_FunctionInfo.insert("memcmp", FunctionInfo("memcmp", 'p', 'p', 'u'));
	m_FunctionInfo.insert("memcpy", FunctionInfo("memcpy", 'p', 'p', 'u'));
	m_FunctionInfo.insert("memmove", FunctionInfo("memmove", 'p', 'p', 'u'));
	m_FunctionInfo.insert("memset", FunctionInfo("memset", 'p', 'c', 'u'));
	m_FunctionInfo.insert("strcat", FunctionInfo("strcat", 's', 's'));
	m_FunctionInfo.insert("strchr", FunctionInfo("strchr", 's', 'c'));
	m_FunctionInfo.insert("strcmp", FunctionInfo("strcmp", 's', 's'));
	m_FunctionInfo.insert("strcoll", FunctionInfo("strcoll", 's', 's'));
	m_FunctionInfo.insert("strcpy", FunctionInfo("strcpy", 'p', 's'));
	m_FunctionInfo.insert("strcspn", FunctionInfo("strcspn", 's', 's'));
	m_FunctionInfo.insert("strerror", FunctionInfo("strerror", 'i'));
	m_FunctionInfo.insert("strlen", FunctionInfo("strlen", 's'));
	m_FunctionInfo.insert("strncat", FunctionInfo("strncat", 's', 's', 'u'));
	m_FunctionInfo.insert("strncmp", FunctionInfo("strncmp", 's', 's', 'u'));
	m_FunctionInfo.insert("strncpy", FunctionInfo("strncpy", 'p', 's', 'u'));
	m_FunctionInfo.insert("strpbrk", FunctionInfo("strpbrk", 's', 's'));
	m_FunctionInfo.insert("strrchr", FunctionInfo("strrchr", 's', 'c'));
	m_FunctionInfo.insert("strspn", FunctionInfo("strspn", 's', 's'));
	m_FunctionInfo.insert("strstr", FunctionInfo("strstr", 's', 's'));
	m_FunctionInfo.insert("strtok", FunctionInfo("strtok", 's', 's'));
	m_FunctionInfo.insert("strxfrm", FunctionInfo("strxfrm", 'p', 's', 'u'));
	m_FunctionInfo.insert("strdup", FunctionInfo("strdup", 's'));
	m_FunctionInfo.insert("strndup", FunctionInfo("strndup", 's', 'u'));
	
	m_FunctionInfo.insert("fopen", FunctionInfo("fopen", 's', 'p'));
	m_FunctionInfo.insert("fclose", FunctionInfo("fclose", 'p'));
	m_FunctionInfo.insert("fread", FunctionInfo("fread", 'p', 'u', 'u', 'p'));
	m_FunctionInfo.insert("fwrite", FunctionInfo("fwrite", 'p', 'u', 'u', 'p'));

	m_FunctionInfo.insert("printf", FunctionInfo("printf", 's'));
	m_FunctionInfo.insert("fputc", FunctionInfo("fputc", 'c', 'p'));
	m_FunctionInfo.insert("fputs", FunctionInfo("fputs", 's', 'p'));
	m_FunctionInfo.insert("putchar", FunctionInfo("putchar", 'c'));
	m_FunctionInfo.insert("puts", FunctionInfo("puts", 'p'));
	m_FunctionInfo.insert("remove", FunctionInfo("remove", 's'));
	m_FunctionInfo.insert("setlocale", FunctionInfo("setlocale", 'i', 's'));
	
	// other functions
	m_FunctionInfo.insert("poll", FunctionInfo("poll", 'p', 'u', 'i'));
	m_FunctionInfo.insert("__libc_start_main", FunctionInfo("__libc_start_main", 'p', 'i', 'p', 'p', 'p', 'p', 'p'));
	m_FunctionInfo.insert("getenv", FunctionInfo("getenv", 's'));
	m_FunctionInfo.insert("setenv", FunctionInfo("setenv", 's', 's', 'i'));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::setupRegisterView(QCatagoryList *catagoryList) {
	// setup the register view
	QTreeWidgetItem *const gpr = catagoryList->addCatagory("General Purpose");
	
	Q_ASSERT(gpr != 0);
	
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // eax
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // ebx
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // ecx
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // edx
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // ebp
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // esp
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // esi
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // edi
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // eip
	m_RegisterViewItems.push_back(catagoryList->addItem(gpr, "")); // eflags
	
	QTreeWidgetItem *const segs = catagoryList->addCatagory("Segments");
	
	Q_ASSERT(segs != 0);
	
	m_RegisterViewItems.push_back(catagoryList->addItem(segs, "")); // cs
	m_RegisterViewItems.push_back(catagoryList->addItem(segs, "")); // ds
	m_RegisterViewItems.push_back(catagoryList->addItem(segs, "")); // es
	m_RegisterViewItems.push_back(catagoryList->addItem(segs, "")); // fs
	m_RegisterViewItems.push_back(catagoryList->addItem(segs, "")); // gs
	m_RegisterViewItems.push_back(catagoryList->addItem(segs, "")); // ss
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QTreeWidgetItem *i386ArchProcessor::getRegisterItem(unsigned int index) {
	QTreeWidgetItem *ret = 0;
	if(index < static_cast<unsigned int>(m_RegisterViewItems.size())) {
		ret = m_RegisterViewItems[index];
	}
	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::doUpdateRegister(QTreeWidgetItem *item, const QString &name, BaseTypes::reg_t value) {
	QString regString;
	// TODO: uses elided text for ascii string?
	if(Debugger::memoryRegions != 0 && Debugger::getASCIIStringAtAddress(value, regString, 4, 128)) {
		item->setText(0, QString().sprintf("%s: %08x ASCII \"%s\"", qPrintable(name), value, qPrintable(regString)));
	} else {
		item->setText(0, QString().sprintf("%s: %08x", qPrintable(name), value));
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::updateRegisterView(const QString &defaultRegionName, const State &state) {
	if(Debugger::memoryRegions != 0) {
		doUpdateRegister(getRegisterItem(0), "EAX", state.eax);
		doUpdateRegister(getRegisterItem(1), "EBX", state.ebx);
		doUpdateRegister(getRegisterItem(2), "ECX", state.ecx);
		doUpdateRegister(getRegisterItem(3), "EDX", state.edx);
		doUpdateRegister(getRegisterItem(4), "EBP", state.ebp);
		doUpdateRegister(getRegisterItem(5), "ESP", state.esp);
		doUpdateRegister(getRegisterItem(6), "ESI", state.esi);
		doUpdateRegister(getRegisterItem(7), "EDI", state.edi);
		
		const QString symname = Debugger::findFunctionSymbol(state.eip, defaultRegionName);
		getRegisterItem(8)->setText(0, QString().sprintf("EIP: %08x <%s>", state.eip, qPrintable(symname)));
		getRegisterItem(9)->setText(0, QString().sprintf("EFLAGS: %08x", state.eflags));

		getRegisterItem(10)->setText(0, QString().sprintf("CS: %04x", state.cs & 0xffff));
		getRegisterItem(11)->setText(0, QString().sprintf("DS: %04x", state.ds & 0xffff));
		getRegisterItem(12)->setText(0, QString().sprintf("ES: %04x", state.es & 0xffff));
		getRegisterItem(13)->setText(0, QString().sprintf("FS: %04x", state.fs & 0xffff));
		getRegisterItem(14)->setText(0, QString().sprintf("GS: %04x", state.gs & 0xffff));
		getRegisterItem(15)->setText(0, QString().sprintf("SS: %04x", state.ss & 0xffff));
		
		
		getRegisterItem(0)->setForeground(0, QBrush((state.eax != m_LastState.eax) ? Qt::red : Qt::black));
		getRegisterItem(1)->setForeground(0, QBrush((state.ebx != m_LastState.ebx) ? Qt::red : Qt::black));
		getRegisterItem(2)->setForeground(0, QBrush((state.ecx != m_LastState.ecx) ? Qt::red : Qt::black));
		getRegisterItem(3)->setForeground(0, QBrush((state.edx != m_LastState.edx) ? Qt::red : Qt::black));
		getRegisterItem(4)->setForeground(0, QBrush((state.ebp != m_LastState.ebp) ? Qt::red : Qt::black));
		getRegisterItem(5)->setForeground(0, QBrush((state.esp != m_LastState.esp) ? Qt::red : Qt::black));
		getRegisterItem(6)->setForeground(0, QBrush((state.esi != m_LastState.esi) ? Qt::red : Qt::black));
		getRegisterItem(7)->setForeground(0, QBrush((state.edi != m_LastState.edi) ? Qt::red : Qt::black));
		
		getRegisterItem(8)->setForeground(0, QBrush((state.eip != m_LastState.eip) ? Qt::red : Qt::black));
		getRegisterItem(9)->setForeground(0, QBrush((state.eflags != m_LastState.eflags) ? Qt::red : Qt::black));

		m_LastState = state;
	} else {
		getRegisterItem(0)->setText(0, "EAX: 00000000");
		getRegisterItem(1)->setText(0, "EBX: 00000000");
		getRegisterItem(2)->setText(0, "ECX: 00000000");
		getRegisterItem(3)->setText(0, "EDX: 00000000");
		getRegisterItem(4)->setText(0, "EBP: 00000000");
		getRegisterItem(5)->setText(0, "ESP: 00000000");
		getRegisterItem(6)->setText(0, "ESI: 00000000");
		getRegisterItem(7)->setText(0, "EDI: 00000000");
		getRegisterItem(8)->setText(0, "EIP: 00000000");
		getRegisterItem(9)->setText(0, "EFLAGS: 00000000");
		getRegisterItem(10)->setText(0, "CS: 0000");
		getRegisterItem(11)->setText(0, "DS: 0000");
		getRegisterItem(12)->setText(0, "ES: 0000");
		getRegisterItem(13)->setText(0, "FS: 0000");
		getRegisterItem(14)->setText(0, "GS: 0000");
		getRegisterItem(15)->setText(0, "SS: 0000");
		
		for(int i = 0; i < 15; ++i) {
			getRegisterItem(7)->setForeground(0, QBrush(Qt::black));
		}
		
		m_LastState = State();
	}
	
	
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QStringList i386ArchProcessor::doFunctionParamterResolution(const QString &symname, const State &state, int offset) {

	QStringList ret;

	// we will always be removing the last 2 chars "+0" from the string as well
	// as chopping the region prefix we like to prepend to symbols
	QString funcName;
	const int indexColon = symname.indexOf("::");

	if(indexColon != -1) {
		funcName = symname.left(symname.length() - 2).mid(indexColon + 2);
	}

	// safe not to check for -1, it means "rest of string" for the mid function
	funcName = funcName.mid(0, funcName.indexOf("@"));
	
	QMap<QString, FunctionInfo>::const_iterator it = m_FunctionInfo.find(funcName);
	if(it != m_FunctionInfo.end()) {
	
		QString functionCall = funcName + "(";
		
		for(int i = 0; i < it->params().size(); ++i) {
			BaseTypes::reg_t arg;
			Debugger::debuggerBase->readBytes(state.esp + i * sizeof(BaseTypes::reg_t) + offset, &arg, sizeof(arg));
			
			switch(it->params()[i].toAscii()) {
			case 's':
				do {
					QString stringParam;
					if(Debugger::getASCIIStringAtAddress(arg, stringParam, 4, INT_MAX)) {
						
						stringParam.replace("\n", "\\n");
						stringParam.replace("\t", "\\t");
					
						functionCall += QString().sprintf("<%#08x> \"%s\"", arg, qPrintable(stringParam));
					} else {
						functionCall += QString().sprintf("%#08x", arg);
					}
				} while(0);
				break;
					
			case 'p':
				functionCall += QString().sprintf("%#08x", arg);
				break;
			case 'c':
				if(safe_ctype<std::isprint>(arg) || arg == ' ') {
					functionCall += QString().sprintf("'%c'", arg);
				} else {
					functionCall += QString().sprintf("'\\x%02x'", arg);
				}
				break;
			case 'u':
				functionCall += QString().sprintf("%u", arg);
				break;
			case 'i':
				functionCall += QString().sprintf("%d", arg);
				break;
			}
			
			if(i + 1 < it->params().size()) {
				functionCall += ", ";
			}
		}
		
		functionCall += ")";
		ret.push_back(functionCall);
	}

	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::analyzeJump(const Operand * const *operands, const State &state, const Instruction &insn, QStringList &ret) {

	Q_UNUSED(operands);

	bool taken = true;
	const bool cf = (state.eflags & 0x0001) != 0;
	const bool pf = (state.eflags & 0x0004) != 0;
	const bool zf = (state.eflags & 0x0040) != 0;
	const bool sf = (state.eflags & 0x0080) != 0;
	const bool df = (state.eflags & 0x0400) != 0;
	const bool of = (state.eflags & 0x0800) != 0;

	if((insn.flagsTested() & insn_carry_set)				&& !cf) taken = false;
	if((insn.flagsTested() & insn_zero_set)					&& !zf) taken = false;
	if((insn.flagsTested() & insn_oflow_set)				&& !of) taken = false;
	if((insn.flagsTested() & insn_dir_set)					&& !df) taken = false;
	if((insn.flagsTested() & insn_sign_set)					&& !sf) taken = false;
	if((insn.flagsTested() & insn_parity_set)				&& !pf) taken = false;
	if((insn.flagsTested() & insn_carry_clear)				&& cf) taken = false;
	if((insn.flagsTested() & insn_zero_clear)				&& zf) taken = false;
	if((insn.flagsTested() & insn_oflow_clear)				&& of) taken = false;
	if((insn.flagsTested() & insn_dir_clear)				&& df) taken = false;
	if((insn.flagsTested() & insn_sign_clear)				&& sf) taken = false;
	if((insn.flagsTested() & insn_parity_clear)				&& pf) taken = false;

	if((insn.flagsTested() & insn_zero_set_or_sign_ne_oflow)&& (!zf && !(sf != of)))	taken = false;
	if((insn.flagsTested() & insn_carry_or_zero_set)		&& !(cf && zf))				taken = false;
	if((insn.flagsTested() & insn_sign_eq_oflow)			&& !(sf == of))				taken = false;
	if((insn.flagsTested() & insn_sign_ne_oflow)			&& (sf == of))				taken = false;

	if(taken) {
		ret.push_back("jump taken");
	} else {
		ret.push_back("jump NOT taken");
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::analyzeReturn(const Operand * const *operands, const State &state, const Instruction &insn, QStringList &ret) {

	Q_UNUSED(operands);
	Q_UNUSED(insn);

	BaseTypes::address_t returnAddress;
	Debugger::debuggerBase->readBytes(state.esp, &returnAddress, sizeof(returnAddress));

	const QString symname = Debugger::findFunctionSymbol(returnAddress, "");
	if(!symname.isEmpty()) {
		ret.push_back(QString().sprintf("Return to %08x <%s>", returnAddress, qPrintable(symname)));
	} else {
		ret.push_back(QString().sprintf("Return to %08x", returnAddress));
	}
}


//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::analyzeCall(const Operand * const *operands, const State &state, const Instruction &insn, QStringList &ret) {
	if(operands[0] != 0) {

		const BaseTypes::address_t effective_address = operands[0]->effectiveAddress(state);
		char tmp[100];

		operands[0]->format(tmp, sizeof(tmp), Debugger::config()->syntax);

		switch(operands[0]->type()) {
		case op_relative_far:
		case op_relative_near:
		case op_register:
			do {
				int offset;
				const QString symname = Debugger::findFunctionSymbol(effective_address, "", &offset);
				if(!symname.isEmpty()) {
					ret.push_back(QString().sprintf("%s = %08x <%s>", tmp, effective_address, qPrintable(symname)));

					if(offset == 0) {
						if(insn.type() == insn_call || insn.type() == insn_callcc) {
							ret += doFunctionParamterResolution(symname, state, 0);
						} else {
							ret += doFunctionParamterResolution(symname, state, 4);
						}
					}

				} else {
					ret.push_back(QString().sprintf("%s = %08x", tmp, effective_address));
				}
			} while(0);
			break;

		case op_expression:
		default:
			do {
				BaseTypes::address_t target;
				const bool ok = Debugger::debuggerBase->readBytes(effective_address, &target, sizeof(target));

				if(ok) {
					int offset;
					const QString symname = Debugger::findFunctionSymbol(target, "", &offset);
					if(!symname.isEmpty()) {
						ret.push_back(QString().sprintf("%s = [%08x] = %08x <%s>", tmp, effective_address, target, qPrintable(symname)));

						if(offset == 0) {
							if(insn.type() == insn_call || insn.type() == insn_callcc) {
								ret += doFunctionParamterResolution(symname, state, 0);
							} else {
								ret += doFunctionParamterResolution(symname, state, 4);
							}
						}

					} else {
						ret.push_back(QString().sprintf("%s = [%08x] = %08x", tmp, effective_address, target));
					}
				} else {
					// could not read from the address
					ret.push_back(QString().sprintf("%s = [%08x] = ?", tmp, effective_address));
				}
			} while(0);
			break;
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::analyzeOperands(const Operand *const *operands, const State &state, const Instruction &insn, QStringList &ret) {

	Q_UNUSED(insn);

	for(int j = 0; j < 3; ++j) {

		if(operands[j] != 0) {

			char tmp[100];
			operands[j]->format(tmp, sizeof(tmp), Debugger::config()->syntax);

			switch(operands[j]->type()) {
			case op_relative_far:
			case op_relative_near:
			case op_register:
				do {
					const BaseTypes::address_t effective_address = operands[j]->effectiveAddress(state);
					ret.push_back(QString().sprintf("%s = %08x", tmp, effective_address));
				} while(0);
				break;
			case op_expression:
				do {
					const BaseTypes::address_t effective_address = operands[j]->effectiveAddress(state);
					BaseTypes::address_t target;

					const bool ok = Debugger::debuggerBase->readBytes(effective_address, &target, sizeof(target));

					if(ok) {
						ret.push_back(QString().sprintf("%s = [%08x] = %08x", tmp, effective_address, target));
					} else {
						ret.push_back(QString().sprintf("%s = [%08x] = ?", tmp, effective_address));
					}
				} while(0);
				break;
			default:
				break;
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QStringList i386ArchProcessor::updateInstructionInfo(BaseTypes::address_t address, const State &state) {

	QStringList ret;
	
	if(Debugger::debuggerBase != 0) {
		uint8_t buffer[MAX_INSN_SIZE];
		Instruction insn;
		
		const bool ok = Debugger::debuggerBase->readBytes(address, buffer, sizeof(buffer));
		if(ok) {
			const int size = insn.disassemble(buffer, sizeof(buffer), address, 0);
			if(size != 0) {

				// fetch the 3 possible operands
				const Operand *const operands[3] = {
					insn.operand_1st(),
					insn.operand_2nd(),
					insn.operand_3rd()
				};

				// figure out the instruction type and display some information about it
				switch(insn.type()) {
				case insn_return:
					analyzeReturn(operands, state, insn, ret);
					break;
				case insn_jcc:
					analyzeJump(operands, state, insn, ret);
					// FALL THROUGH!
				case insn_jmp:
				case insn_call:
				case insn_callcc:
					analyzeCall(operands, state, insn, ret);
					break;
				default:
					analyzeOperands(operands, state, insn, ret);
					break;
				}
			}
		}
	}
	
	// eliminate duplicates
	ret = QStringList::fromSet(ret.toSet());
	
	return ret;
}
